#!/bin/bash -euET

#  Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew

CLEANEXIT=0
trap '
SAVEDSTATUS=$?
set +x
if [ "${CLEANEXIT:-0}" -ne 1 ]
then
	echo "${0##*/}: FATAL Unexpected return value of ${SAVEDSTATUS:-?}, line number ${ERR_LINENO:-?}, quitting" 1>&2
	exit $SAVEDSTATUS
fi' EXIT

trap 'ERR_LINENO=$LINENO' ERR

set -o pipefail

#-----

more << 'EOF'
This is the Host Factory pre-build setup script.

On a client/target/remote Ubuntu 14 managed with "hfdist", only the
"/opt/hf" portion of a Host Factory binary distribution and the
apt-get package "openssh-server" is necessary.  Any distribution of
perl 5 which has File::Path should work.  To get the binary
distribution onto the remote, see the "hf-bootstrap-*" commands.

However, this script sets up building the Host Factory source with all
its features active, which requires latex, mysql, virtualbox, and
vagrant among others.  If apt-get asks you to choose between the local
version of a configuration file or the package maintainer's version,
pick the package's version.

Building fewer features without big packages installed is possible
using only these packages and "make less", "make installless":

	mercurial
	perl-doc
	rdb
	PATH for rdb ahead of system path to shadow "column"
	source code checked into Mercurial
	install directories /opt/hf /opt/hfbulk /opt/hfsite

This package adds the ability to contact "hfdist -S", the remote side
of file transfers, through ssh:

	openssh-server

This package adds the graphical history command "hg view" in
Mercurial:

	tk

These packages enable "make", and allow building the book:

	latex, 1 Gig download
	inkscape
	tree
	graphviz

This package adds usability of config file generation and full "make
test":

	mysql

These packages add the vagrant/virtualbox demo script
"blah/XXX/demo.sh":

	virtualbox
	vagrant

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

more << 'EOF'

Having the source under Mercurial version control is necessary to
build, as the version id gets placed into binaries.  Mercurial will be
downloaded and set up.

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install mercurial)

if [ ! -f ~/.hgrc ]
then
	cat << 'EOF' > ~/.hgrc
[ui]
editor=emacs
username=Your Name <youremail@example.com>

[extensions]
hgk =
EOF

	more << 'EOF'

Edit ~/.hgrc to enter your identity information.

EOF
	echo "Press enter to continue or ^C to cancel"
	read -e -p '> ' -i '' JUNK
fi

more << 'EOF'

The source code needs to be in a Mercurial repository to build.  If it
is not, do this in the source directory:

	$ hg init
	$ hg add
	$ hg com -m 'Initial checkin'

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

if [ ! -d /opt/rdb-2.1/bin ]
then
	more << 'EOF'

Now installing the rdb database scripts to manipulate tab-separated
data files in a stream fashion.

EOF
	echo "Press enter to continue or ^C to cancel"
	read -e -p '> ' -i '' JUNK

	build-site/apt-get-packages/precise-12/INSTALL-RDB.sh
fi

if ! echo $PATH | grep -q '/rdb-2\.1/'
then
	more << 'EOF'

rdb needs to be ahead of the system path to shadow the "column"
command.  Edit ~/.bashrc or similar to update PATH like:

	export PATH=/opt/rdb-2.1/bin:/opt/hf/hf-x.y/bin:$PATH

To incorporate the PATH change after you finish running this script,
do "exec bash" before trying "make".

EOF
	echo "Press enter to continue or ^C to cancel"
	read -e -p '> ' -i '' JUNK
fi

more << 'EOF'

Next, perl-doc will be installed to produce man pages from pod.

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install perl-doc)

if [ ! \( -d /opt/hf -a -d /opt/hfbulk -a -d /opt/hfsite \) ]
then
	more << 'EOF'

The installation directories /opt/hf, /opt/hfbulk, and /opt/hfsite/hfgen.out
will be created.

EOF
	echo "Press enter to continue or ^C to cancel"
	read -e -p '> ' -i '' JUNK

	(set -x
		sudo mkdir -p /opt/hf /opt/hfbulk /opt/hfsite/hfgen.out
		sudo chown -R ${USER}.${USER} /opt/hf /opt/hfbulk /opt/hfsite)
fi

more << 'EOF'

At this point you can build all but the book with "make less" and
install it with "make installless".

Do you need to "exec bash" to incorporate a PATH change?

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

more << 'EOF'

Incoming ssh is not necessary to build, but it is to use between
hosts.  The server/source/local needs outgoing ssh, and the
client/target/remote needs incoming ssh:

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install openssh-client openssh-server openssh-blacklist openssh-blacklist-extra)

more << 'EOF'

tk is not necessary to build, but it is so hg view will work:

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install tk8.5)

more << 'EOF'

expect is not necessary to build, but it is so tests/test-oh will work:

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install expect)

more << 'EOF'

Devel::NYTProf is not necessary to build, but it is so make profile
will work:

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install libdevel-nytprof-perl)

more << 'EOF'

Latex is not necessary to build bash and perl programs, but it is to
build the book.  This latex distribution is a gigabyte of download.

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install texlive texlive-doc-en dvipdfmx lmodern tipa prosper latex-beamer latex-xcolor texpower texlive-generic-extra cm-super texlive-latex-extra texlive-full
	sudo texconfig-sys dvips paper letter
	sudo texconfig-sys pdftex paper letter

	more << 'EOF'

If prompted, overwrite with maintainer's version

EOF
	echo "Press enter to continue or ^C to cancel"
	read -e -p '> ' -i '' JUNK

	sudo texconfig-sys dvipdfm paper letter
	sudo texconfig-sys xdvi paper us
	sudo texconfig-sys mode ljfour)

more << 'EOF'

graphviz, tree, and inkscape programs for the book

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

(set -x
	sudo apt-get install graphviz graphviz-doc
	sudo apt-get install tree
	sudo apt-get install inkscape python-lxml-dbg python-numpy-doc python-numpy-dbg python-uniconvertor-dbg libsvg-perl python-reportlab-doc)

if [ ! -f ~/.hfgen-password ]
then
	more << 'EOF'

Mysql is not necessary to build, but it is necessary to test and use
fully.  Needed is a working mysql with 'create table' privilege for
the accounts that will do the Host Factory db work, and a user 'hfgen'
with read only access and a password in a dot file.  If you are
setting up mysql from scratch, this script from the source directory
will load the apt-get packages and do the setup:

	build-site/apt-get-packages/trusty-14/INSTALL-MYSQL.sh

Go run that script, then come back here.  If you already have mysql,
then read the script and do whatever is appropriate for your site.

After you've completed PREBUILD.sh and built and installed the Host
Factory source, but before you proceed past the prompt for the
database tests in "make test", the mysql database "hfexample1" needs
to have known contents for the database related tests to work.  This
command loads it:

	hf-reloaddb

EOF
	echo "Press enter to continue or ^C to cancel"
	read -e -p '> ' -i '' JUNK
fi

more << 'EOF'

Vagrant and virtualbox are not necessary to build, but they are
necessary to run the vagrant demo "demo/XXX.sh".  They are not loaded
here because you may be using the upstream versions due to bug fixes.
If you want to load the apt-get versions, do:

	sudo apt-get install virtualbox vagrant

EOF
echo "Press enter to continue or ^C to cancel"
read -e -p '> ' -i '' JUNK

### (set -x
###	sudo apt-get install virtualbox
###	sudo apt-get install vagrant)

more << 'EOF'

Do you need to "exec bash" to incorporate a PATH change?

"make test" comes after "make install" because perl "use lib" uses
perl modules from the install location during "make test".

Do you later need to "hf-reloaddb" to load db "hfexample1"?

EOF

#-----

CLEANEXIT=1
exit 0
