#!/bin/bash -euET

#  Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew

#
#  Usage:
#
#	demo.sh [-y]
#
#	-y  Don't stop at prompts to continue
#

CLEANEXIT=0
trap '
SAVEDSTATUS=$?
set +x
if [ "${CLEANEXIT:-0}" -ne 1 ]
then
	echo "${0##*/}: FATAL Unexpected return value of ${SAVEDSTATUS:-?}, line number ${ERR_LINENO:-?}, quitting" 1>&2
	exit $SAVEDSTATUS
fi' EXIT

trap 'ERR_LINENO=$LINENO' ERR

set -o pipefail

#-----

NOSTOP="${1:-}" ; shift || :

if ! type -p vagrant > /dev/null
then
	echo "${0##*/}: FATAL vagrant not installed, quitting" 1>&2
	exit 1
fi

if ! type -p virtualbox > /dev/null
then
	echo "${0##*/}: FATAL virtualbox not installed, quitting" 1>&2
	exit 1
fi

if [ `id -u` -ne 0 ]
then
	echo "${0##*/}: FATAL must be run as root, quitting" 1>&2
	exit 1
fi

# Files used in this script are found in the same directory as this script is in
# pwd -P because I want the real directory, not the symlink on the path used to get there
EXEDIRNAME=`dirname $0`
EXEDIR=`(cd "$EXEDIRNAME" ; pwd -P)`

function pauseprompt() {

	echo -n "$1"
	if [ "$NOSTOP" != "-y" ]
	then
		echo ", press enter to continue or ^C"
		read -e -p '> ' -i '' JUNK
	else
		echo
	fi
}

#-----

pauseprompt '			       Host Factory demo
-------------------------------------------------------------------------------

 1. Clean up from previous runs
 2. Create two Ubuntu Linux hosts on VirtualBox with vagrant
 3. Enable incoming ssh as root on both hosts
 4. Install Host Factory on both hosts
 5. Use hfdist to pull and capture filesystem contents of host1
 6. Check captured host1 filesystem contents into oh version control
 7. Adjust captured files for per-host differences:
	/var/log files are emptied
	/etc/hostname files are generated
	/dev/disk/by-id/{ata-,scsi-SATA_}VBOX_HARDDISK* symlinks are captured
 8. Check per-host changes into oh version control
 9. Demonstrate hfdist push to original capture host is reasonable
10. Demonstrate hfdist push to other host is reasonable
11. Install tcsh on original capture host using apt-get package system
12. Use hfdist to pull changes from tcsh install
13. Check tcsh changes into oh version control
14. Install tcsh on other host using apt-get package system
15. Push to other host to fix unwanted differences left by package system
16. Add rootkit
17. Detect rootkit
18. Clean up rootkit
19. Halt both hosts

== (1 of 19) About to clean up from previous runs'

( set -x
	for i in host1 host2
	do
		if [ -d "vagrant/$i" ]
		then
			cd "vagrant/$i"
			vagrant halt -f || :
			vagrant destroy -f || :
			cd ../..
		fi
	done
	rm -rf vagrant image gen per-host
)

pauseprompt '== (2 of 19) About to create the virtual machines'

( set -x
	mkdir vagrant
	for i in host1 host2
	do
		cd vagrant
		mkdir $i
		cd $i
		vagrant init hashicorp/precise64
		vagrant up
		cd ../..
	done
)

pauseprompt '== (3 of 19) About to enable incoming ssh as root'

( set -x
	for i in host1 host2
	do
		cd vagrant/$i
		echo -e 'sudo -i\nmkdir /root/.ssh\nchmod 700 /root/.ssh\ncp /home/vagrant/.ssh/authorized_keys /root/.ssh\n' | vagrant ssh -c bash
		cd ../..
	done
)

pauseprompt '== (4 of 19) About to install Host Factory on the machines'

( set -x
	for i in host1 host2
	do
		hf-bootstrap-push-bash | $EXEDIR/vagrant-ssh $i
	done
)

pauseprompt '== (5 of 19) About to initially capture machine files'

( set -x
	mkdir image
	cd image
	oh init
	cd ..
	
	time hfdist -v 6 -wd -r -l $EXEDIR/initial-pull.hflayers -P '["-F", "vagrant/host1/ssh-config"]' root@default
)

pauseprompt '== (6 of 19) About to version control the initial capture'

( set -x
	time (cd image ; oh snapshot initial pull)
)

pauseprompt '== (7 of 19) About to adjust captured files for per-host differences'

( set -x
	mkdir gen
	cd gen
	oh init
	for i in host1 host2
	do
		mkdir oh/$i
	done
	cd ..
	
	for i in host1 host2
	do
		mkdir -p gen/oh/$i/etc
		echo $i > gen/oh/$i/etc/hostname
	done
	
	find image/oh/{var/log,etc/hostname} -type f -print | \
		while read line ; do cat /dev/null > $line ; done
	
	mkdir per-host
	cd per-host
	oh init
	for i in host1 host2
	do
		mkdir oh/$i
	done
	cd ..
	
	(cd image/oh ; tar -cf - dev/disk/by-id/ata-VBOX_HARDDISK* dev/disk/by-id/scsi-SATA_VBOX_HARDDISK*) | \
		(cd per-host/oh/host1 ; tar -xvpf -)
	
	find image/oh/dev/disk/by-id/ata-VBOX_HARDDISK* image/oh/dev/disk/by-id/scsi-SATA_VBOX_HARDDISK* -print | \
		xargs rm
	
	echo 'cd / ; tar -cf - dev/disk/by-id/ata-VBOX_HARDDISK* dev/disk/by-id/scsi-SATA_VBOX_HARDDISK*' | \
		$EXEDIR/vagrant-ssh host2 bash | (cd per-host/oh/host2 ; tar -xvpf -)
)

pauseprompt '== (8 of 19) About to version control the per-host changes'

( set -x
	for i in image gen per-host
	do
		time (cd $i ; oh snapshot after per-host adjustments)
	done
)

pauseprompt '== (9 of 19) About to show that push results to original capture host are reasonable'

( set -x
	export LAYERS_HOST=host1 ; time hfdist -v 6 -l $EXEDIR/push.hflayers -P '["-F", "vagrant/'$LAYERS_HOST'/ssh-config"]' root@default
)

pauseprompt '== (10 of 19) About to show that push results to other host are reasonable'

( set -x
	export LAYERS_HOST=host2 ; time hfdist -v 6 -l $EXEDIR/push.hflayers -P '["-F", "vagrant/'$LAYERS_HOST'/ssh-config"]' root@default
)

pauseprompt '== (11 of 19) About to install tcsh on original capture host'

( set -x
	$EXEDIR/vagrant-ssh host1 apt-get -y install tcsh
)

pauseprompt '== (12 of 19) About to pull changes from tcsh install'

( set -x
	export LAYERS_HOST=host1 ; time hfdist -v 6 -wd -srk -l $EXEDIR/push.hflayers -P '["-F", "vagrant/'$LAYERS_HOST'/ssh-config"]' root@default
)

pauseprompt '== (13 of 19) About to version control tcsh changes'

( set -x
	time (cd image ; oh snapshot after adding tcsh)
)

pauseprompt '== (14 of 19) About to install tcsh on other host with package system'

( set -x
	$EXEDIR/vagrant-ssh host2 apt-get -y install tcsh
)

pauseprompt '== (15 of 19) About to hfdist over unwanted differences package system left between the two hosts'

( set -x
	export LAYERS_HOST=host2 ; time hfdist -v 6 -wd -l $EXEDIR/push.hflayers -P '["-F", "vagrant/'$LAYERS_HOST'/ssh-config"]' root@default
)

pauseprompt '== (16 of 19) Add rootkit'

( set -x
	echo 'set -x ; touch -r /usr/bin /tmp/ub ; cp /bin/sh /usr/bin/hpasswd ; chown root.root /usr/bin/hpasswd ; chmod 4755 /usr/bin/hpasswd ; touch -r /usr/bin/gpasswd /usr/bin/hpasswd ; touch -r /tmp/ub /usr/bin ; rm /tmp/ub' | $EXEDIR/vagrant-ssh host2
)

pauseprompt '== (17 of 19) Detect rootkit'

( set -x
	export LAYERS_HOST=host2 ; time hfdist -v 3 -l $EXEDIR/push.hflayers -P '["-F", "vagrant/'$LAYERS_HOST'/ssh-config"]' root@default
)

pauseprompt '== (18 of 19) Clean up rootkit'

( set -x
	export LAYERS_HOST=host2 ; time hfdist -v 3 -wd -l $EXEDIR/push.hflayers -P '["-F", "vagrant/'$LAYERS_HOST'/ssh-config"]' root@default
)

pauseprompt '== (19 of 19) About to halt hosts'

( set -x
	for i in host1 host2
	do
		cd vagrant/$i
		vagrant halt
		cd ../..
	done
)

#-----

CLEANEXIT=1
exit 0
