#  Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew

package HF::FTGroveler;
use strict;

use Digest::MD5;

sub new {

	my $proto = shift;
	my $startdir = shift;
	my $ignore = shift;
	my $dochecksums = shift;

	my $class = ref ($proto) || $proto;

	my $self = {

		'array'		=> [ ],
		'currpath'	=> [ ],
		'descend'	=> [ ],
		'dirindex'	=> [ ],
		'numdirents'	=> [ ],
		'ignore'	=> $ignore,
		'index'		=> 0,
		'pathindex'	=> -1,
		'startdir'	=> $startdir,
		'encstartdir'	=> HF::Encode->encode ($startdir),
		'striplen'	=> 0,
		'dochecksums'	=> $dochecksums,
		'md5handle'	=> Digest::MD5->new,
		'isdone'	=> 0
	};

# NOTE: index and pathindex point to LAST (CURRENT) element!

	$self->{'array'}[$self->{'index'}] = [ $self->{'startdir'} ];
	$self->{'descend'}[$self->{'index'}] = 0;
	$self->{'dirindex'}[$self->{'index'}] = 0;
	$self->{'numdirents'}[$self->{'index'}] = 1;

# XXX assume reasonable input path, no dot dot inside, for instance

	if ($startdir eq '/') {
		$self->{'striplen'} = 1;
	} else {
		if ($startdir =~ /\/$/) {
			$self->{'striplen'} = length ($startdir);
		} else {
			$self->{'striplen'} = length ($startdir) + 1;
		}
	}

	bless ($self, $class);

	return ($self);
}

sub next {

	my $self = shift;

	my $DH;
	my $dir;
	my $path;
	my $str;
	my $reshash;
	my $looklink;
	my $pathnoslash;

	if ($self->{'isdone'} == 1) {
		HF::Report->report (4, 1, "FATAL called after end of walk");
	}

	$reshash->{'startdir'} = $self->{'startdir'};
	$reshash->{'encstartdir'} = $self->{'encstartdir'};

	#
	# If index is < 0, assume we're being asked to walk off the end
	# of our tree; don't do it
	#
	if ($self->{'index'} < 0) {
		$self->{'isdone'} = 1;
		return ($reshash);
	}

	#
	# descend into a directory
	#
	if ($self->{'descend'}[$self->{'index'}]) {
		$self->{'currpath'}[++$self->{'pathindex'}] =
			$self->{'array'}[$self->{'index'}]->[$self->{'dirindex'}[$self->{'index'}]];
		$self->{'descend'}[$self->{'index'}] = 0;

		#
		# Open the directory
		#
		$path = $self->makePath;

		#  Lessons learned:
		#
		#  1) don't opendir a non-dir.
		#
		#  2) don't opendir a path foo/, because if foo is a symlink pointing to a dir it will opendir
		#
		#  3) don't use opendir as a test to find dirs.  opendir will open a symlink pointing to a dir.
		#
		#  4) the filesystem will return the update if you replace dir foo with symlink foo
		#
		#  5) without a way to tell FTG that its cache has been invalidated by stuff done outside,
		#     there needs to be a stat just before every opendir to notice type changes and deletions.
		#     some "errors" from data changing underfoot are to be expected.

		$pathnoslash = $path;
		if (length ($pathnoslash) >= 2) {
			$pathnoslash =~ s|/$||;
		}

		($looklink->{'dev'}, $looklink->{'inode'}, $looklink->{'mode'},
			$looklink->{'links'}, $looklink->{'owner'}, $looklink->{'group'},
			$looklink->{'rdev'}, $looklink->{'size'}, $looklink->{'atime'},
			$looklink->{'mtime'}, $looklink->{'ctime'},
			$looklink->{'blksize'}, $looklink->{'blocks'}) = lstat ($pathnoslash);

		$looklink->{'type'} = $looklink->{'mode'} & 0170000;
		$looklink->{'perms'} = $looklink->{'mode'} & 07777;

		if ((($looklink->{'mode'} & 0170000) != 0040000) || (!(opendir ($DH, $pathnoslash)))) {

			# perhaps dir was deleted by target_compare before we could scan it
			HF::Report->report (2, 2, "FAILURE 2 can't opendir '" . HF::Encode->encode ($pathnoslash) . "', error '$!'", 'fail_layers');
			$self->popDir;
			++$self->{'dirindex'}[$self->{'index'}];
			return ($self->next);
		}

		#
		# Read it and sort it
		#
		@$dir = sort (grep (!/^\.\.?$/, readdir ($DH)));
		if (!(closedir ($DH))) {
			HF::Report->report (2, 2, "FAILURE can't closedir '" . HF::Encode->encode ($path) . "', error '$!'", 'fail_layers');
		}
		if (!($dir)) {
			HF::Report->report (2, 2, "FAILURE can't readdir '" . HF::Encode->encode ($path) . "', error '$!'", 'fail_layers');
			$self->popDir;
			++$self->{'dirindex'}[$self->{'index'}];
			return ($self->next);
		}

		#
		# Diddle a bunch of pointers
		#
		++$self->{'index'};
		$self->{'array'}[$self->{'index'}] = $dir;
		$self->{'dirindex'}[$self->{'index'}] = 0;
		$self->{'numdirents'}[$self->{'index'}] = scalar (@$dir);
		return ($self->next);
	}

	#
	# Return to our parent directory
	#
	if ($self->{'dirindex'}[$self->{'index'}] == $self->{'numdirents'}[$self->{'index'}]) {
		$self->popDir;
		--$self->{'index'};
		if ($self->{'index'} < 0) {
			$self->{'isdone'} = 1;
			return ($reshash);
		} else {
			++$self->{'dirindex'}[$self->{'index'}];
			$self->{'descend'}[$self->{'index'}] = 0;
			return ($self->next);
		}
	}

	#
	# Process the next entry in the directory
	#
	$path = $self->makePath . $self->{'array'}[$self->{'index'}]->[$self->{'dirindex'}[$self->{'index'}]];

	$reshash->{'path'} = $path;
	$reshash->{'encpath'} = HF::Encode->encode ($path);

	$str = $path;
	$str = substr ($path, $self->{'striplen'});

	($reshash->{'dev'}, $reshash->{'inode'}, $reshash->{'mode'},
		$reshash->{'links'}, $reshash->{'owner'}, $reshash->{'group'},
		$reshash->{'rdev'}, $reshash->{'size'}, $reshash->{'atime'},
		$reshash->{'mtime'}, $reshash->{'ctime'},
		$reshash->{'blksize'}, $reshash->{'blocks'}) = lstat ($path);

	if (!(defined ($reshash->{'dev'}))) {
		HF::Report->report (2, 2, "FAILURE 3 can't lstat '" . $reshash->{'encpath'} . "', error '$!'", 'fail_layers');

		if ($self->{'pathindex'} < 0) {
			return ($reshash);
		} else {
			return ($self->next);
		}
	}

	$reshash->{'type'} = $reshash->{'mode'} & 0170000;
	$reshash->{'perms'} = $reshash->{'mode'} & 07777;

	if ($str ne '') {
		$reshash->{'name'} = $str;
		$reshash->{'encname'} = HF::Encode->encode ($str);
	} else {
		if ($reshash->{'type'} == 0040000) {
			#  Name for top-level dir of walk is returned as dot, regardless of
			#  actual name of dir
			$str = '.';
			$reshash->{'name'} = $str;
			$reshash->{'encname'} = HF::Encode->encode ($str);
		} else {
			#  If walk startdir is non-dir, set name to basename of walk
			$str = $path;
			$str =~s,^.*/,,;
			$reshash->{'name'} = $str;
			$reshash->{'encname'} = HF::Encode->encode ($str);
		}
	}

	if (defined ($self->{'ignore'}) && ($str =~ /$self->{'ignore'}/)) {
		HF::Report->report (0, 7, "PROGRESS skipping '" . $reshash->{'encname'} . "'");
		if (defined ($HF::Globals::outputplbig)) {
			push (@{$HF::Globals::wasskipped}, $reshash->{'name'});
		}
		++$self->{'dirindex'}[$self->{'index'}];
		return ($self->next);
	}

	$reshash->{'target'} = undef;
	$reshash->{'enctarget'} = undef;

	$reshash->{'checksum'} = undef;

	if (($reshash->{'type'} == 0100000) && ($self->{'dochecksums'} == 1)) {
		$reshash->{'checksum'} = $self->getchecksum ($path);
	}

	if ($reshash->{'type'} == 0120000) {

		++$self->{'dirindex'}[$self->{'index'}];
		$reshash->{'target'} = readlink ($path);
		if (!(defined ($reshash->{'target'}))) {
			HF::Report->report (2, 2, "FAILURE 2 can't readlink '" . $reshash->{'encpath'} . "', error '$!'", 'fail_layers');
			return ($self->next);
		}
		$reshash->{'enctarget'} = HF::Encode->encode ($reshash->{'target'});
		return ($reshash);

	} elsif ($reshash->{'type'} != 0040000) {

		++$self->{'dirindex'}[$self->{'index'}];
		return ($reshash);

	} else {

		$self->{'descend'}[$self->{'index'}] = 1;
		return ($reshash);
	}
}

#
# Short, but used several places
#
sub popDir {

	my $self = shift;

	--$self->{'pathindex'};
	--$#{$self->{'currpath'}};
}

#
# Ibid.
#
sub makePath {

	my $self = shift;

# XXX clean up path creation logic

	if ($self->{'pathindex'} < 0) {
		return ('');
	} elsif ($self->{'pathindex'} > 0) {
		if ($self->{'currpath'}[0] eq '/') {
			return (substr (join ('/', @{$self->{'currpath'}}) . '/', 1));
		} else {
			return (join ('/', @{$self->{'currpath'}}) . '/');
		}
	} else {
		if ($self->{'currpath'}[$self->{'pathindex'}] eq '/') {
			return ($self->{'currpath'}[$self->{'pathindex'}]);
		} else {
			return ($self->{'currpath'}[$self->{'pathindex'}] . '/');
		}
	}
}

sub getchecksum {

	my $self = shift;
	my $filename = shift;

	my $MD5;

	$self->{'md5handle'}->reset;
	if (open ($MD5, "<", $filename)) {
		$self->{'md5handle'}->addfile ($MD5);
		close ($MD5) || HF::Report->report (3, 1, "FATAL close of md5 fh for filename '$filename' failed, error '$!'");
		return ($self->{'md5handle'}->hexdigest);
	} else {
		HF::Report->report (2, 2, "FAILURE 4 can't open '" . HF::Encode->encode ($filename) . "' for checksum, error '$!'", 'fail_layers');
		return (undef);
	}
}

1;

__END__

=head1 NOTES

May still confuse a symlink to a dir with a real dir.

=cut
