#  Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew

package HF::FTWstream;
use strict;

use Digest::MD5;

sub new {

	my $proto = shift;
	my $startdir = shift;
	my $ignore = shift;
	my $dochecksums = shift;

	my $class = ref ($proto) || $proto;

	my $self = {

		'ignore'	=> $ignore,
		'dochecksums'	=> $dochecksums,

		'md5handle'	=> Digest::MD5->new,
		'isdone'	=> 0
	};

# XXX assume reasonable input path, no dot dot inside, for instance

# Should work for all of:
#	.
#	/
#	/foo
#	/foo/
#	foo
#	foo/
#	./foo
#	./foo/

	if ($startdir eq "" ) {
		HF::Report->report (4, 1, "FATAL startdir parameter zero-length");
	} elsif ($startdir eq "/") {
		# remove / from startdir, accept / from joining paths as its replacement
		$self->{'striplen'} = 1;
	} else {
		$startdir =~ s|/$||;
		# remove / from joining paths
		$self->{'striplen'} = length ($startdir) + 1;
	}

	$self->{'startdir'} = $startdir;
	$self->{'encstartdir'} = HF::Encode->encode ($startdir);

	$self->{'level'} = 0;
	$self->{'name'} = [ $startdir ];
	$self->{'state'} = [ 1 ];
	$self->{'dirlist'} = [ ];

	bless ($self, $class);

	return ($self);
}

sub next {

	my $self = shift;

	my $entry;
	my $FTWSTREAM;
	my @lstatfields;
	my $ignorename;

	if ($self->{'isdone'} == 1) {
		HF::Report->report (4, 1, "FATAL called after end of walk");
	}

	#  Exits when returning output to user
	for (;;) {

		#  Finished walk, no more entries to return
		if ($self->{'level'} == -1) {
			$self->{'isdone'} = 1;
			$self->{'res'} = { };
			return ($self->{'res'});
		}

		if ($self->{'state'}[$self->{'level'}] == 1) {

			if ((defined ($self->{'ignore'})) && ($ignorename = substr ($self->{'name'}[$self->{'level'}], $self->{'striplen'})) && ($ignorename =~ /$self->{'ignore'}/)) {
				HF::Report->report (0, 7, "PROGRESS skipping '" . HF::Encode->encode ($ignorename) . "'");
				if (defined ($HF::Globals::outputplbig)) {
					push (@{$HF::Globals::wasskipped}, $ignorename);
				}
				$self->{'level'}--;
			} else {
				$self->{'state'}[$self->{'level'}] = 2;
			}

		} elsif ($self->{'state'}[$self->{'level'}] == 2) {

			$self->{'res'} = $self->lstat_work;

			if (!(defined ($self->{'res'}{'inode'}))) {
				$self->{'level'}--;
			} else {

				if ($self->{'res'}{'type'} == 0040000) {
					$self->{'state'}[$self->{'level'}] = 3;
					return ($self->{'res'});
				} else {
					$self->{'level'}--;
					return ($self->{'res'});
				}
			}

		} elsif ($self->{'state'}[$self->{'level'}] == 3) {

			# This entry might be a different type now since the intervening return.

			@lstatfields = lstat ($self->{'name'}[$self->{'level'}]);

			if ((($lstatfields[2]) & 0170000) == 0040000) {

				if (opendir ($FTWSTREAM, $self->{'name'}[$self->{'level'}])) {
					# Filter out dot and dotdot.
					@{$self->{'dirlist'}[$self->{'level'}]} = sort (grep (!/^\.\.?$/, readdir ($FTWSTREAM)));
					closedir ($FTWSTREAM);
					$self->{'state'}[$self->{'level'}] = 4;
				} else {
					# If it's a dir that's unreadable, perhaps due to permissions,
					# skip it like it doesn't exist.
					HF::Report->report (2, 2, "FAILURE 1 can't opendir '" . $self->{'name'}[$self->{'level'}] . "', error '$!'", 'fail_layers');
					$self->{'level'}--;
				}
			} else {
				# Not our error, this dir was changed or removed behind my back
				$self->{'level'}--;
			}

		} elsif ($self->{'state'}[$self->{'level'}] == 4) {

			$entry = shift (@{$self->{'dirlist'}[$self->{'level'}]});

			if ($entry ne "") {
				if ($self->{'name'}[$self->{'level'}] =~ /\/$/) {
					# for instance when startdir is /
					$self->{'name'}[$self->{'level'} + 1] = $self->{'name'}[$self->{'level'}] . $entry;
				} else {
					$self->{'name'}[$self->{'level'} + 1] = $self->{'name'}[$self->{'level'}] . '/' . $entry;
				}
				$self->{'state'}[++$self->{'level'}] = 1;
			} else {
				$self->{'level'}--;
			}

		} else {
			HF::Report->report (4, 0, "FATAL unknown state " . $self->{'state'}[$self->{'level'}]);
		}
	}
	HF::Report->report (4, 0, "FATAL should never reach this statement");
}

sub lstat_work {

	my $self = shift;

	my $res;

	$res = {
		'startdir' => $self->{'startdir'},
		'encstartdir' => HF::Encode->encode ($self->{'startdir'}),
		'path' => $self->{'name'}[$self->{'level'}],
		'encpath' => HF::Encode->encode ($self->{'name'}[$self->{'level'}])
	};

	($res->{'dev'}, $res->{'inode'}, $res->{'mode'},
		$res->{'links'}, $res->{'owner'}, $res->{'group'},
		$res->{'rdev'}, $res->{'size'}, $res->{'atime'},
		$res->{'mtime'}, $res->{'ctime'},
		$res->{'blksize'}, $res->{'blocks'}) = lstat ($self->{'name'}[$self->{'level'}]);

	if (!(defined ($res->{'inode'}))) {
		return ($res);
	} else {
		$res->{'type'} = $res->{'mode'} & 0170000;
		$res->{'perms'} = $res->{'mode'} & 07777;

		$res->{'name'} = substr ($res->{'path'}, $self->{'striplen'});
		if ($res->{'name'} eq '') {
			if ($res->{'type'} == 0040000) {
				$res->{'name'} = ".";
			} else {
				$res->{'name'} = $res->{'path'};
				$res->{'name'} =~ s|^.*/||;
			}
		}
		$res->{'encname'} = HF::Encode->encode ($res->{'name'});

		if ($res->{'type'} == 0120000) {
			$res->{'target'} = readlink ($self->{'name'}[$self->{'level'}]);
			if (!(defined ($res->{'target'}))) {
				HF::Report->report (2, 2, "FAILURE 1 can't readlink '" . $self->{'name'}[$self->{'level'}] . "', error '$!'", 'fail_layers');
			}
			$res->{'enctarget'} = HF::Encode->encode ($res->{'target'});

		}

		if (($res->{'type'} == 0100000) && ($self->{'dochecksums'} == 1)) {
			$res->{'checksum'} = $self->getchecksum ($self->{'name'}[$self->{'level'}]);
		} else {
			$res->{'checksum'} = undef;
		}

		return ($res);
	}
}

sub stattohash {

	my $self = shift;
	my $path = shift;

	my $reshash;
	my $str;

	$reshash->{'name'} = $path;
	$reshash->{'encname'} = HF::Encode->encode ($path);

	$reshash->{'path'} = $reshash->{'name'};
	$reshash->{'encpath'} = $reshash->{'encname'};

	($reshash->{'dev'}, $reshash->{'inode'}, $reshash->{'mode'},
		$reshash->{'links'}, $reshash->{'owner'}, $reshash->{'group'},
		$reshash->{'rdev'}, $reshash->{'size'}, $reshash->{'atime'},
		$reshash->{'mtime'}, $reshash->{'ctime'},
		$reshash->{'blksize'}, $reshash->{'blocks'}) = lstat ($path);

	if (!(defined ($reshash->{'dev'}))) {
		HF::Report->report (2, 2, "FAILURE 2 can't lstat '" . $reshash->{'encpath'} . "', error '$!'", 'fail_layers');
	}

	$reshash->{'type'} = $reshash->{'mode'} & 0170000;
	$reshash->{'perms'} = $reshash->{'mode'} & 07777;

	$reshash->{'target'} = undef;
	$reshash->{'enctarget'} = undef;

	$reshash->{'checksum'} = undef;

	return ($reshash);
}

sub getchecksum {

	my $self = shift;
	my $filename = shift;

	my $MD5;

	$self->{'md5handle'}->reset;
	if (open ($MD5, "<", $filename)) {
		$self->{'md5handle'}->addfile ($MD5);
		close ($MD5) || HF::Report->report (4, 1, "FATAL close of md5 fh for filename '$filename' failed, error '$!'");
		return ($self->{'md5handle'}->hexdigest);
	} else {
		HF::Report->report (2, 2, "FAILURE 3 can't open '" . HF::Encode->encode ($filename) . "' for checksum, error '$!'", 'fail_layers');
		return (undef);
	}
}

1;

__END__

=head1 NOTES

Tolerant of mutations made to the portion of the file tree walk it
hasn't returned yet.  Double-checks its own prefetch queue entries to
catch data made stale by Harmonize.

=cut
