#  Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew

package HF::GLOBstream;
use strict;

use Digest::MD5;

sub new {

	my $proto = shift;
	my $globexp = shift;
	my $ignore = shift;
	my $dochecksums = shift;

	my $class = ref ($proto) || $proto;

	my $self = {

		'ignore'	=> $ignore,
		'dochecksums'	=> $dochecksums,

		'globresult'	=> undef,
		'firsttime'	=> 1,

		'md5handle'	=> Digest::MD5->new,
		'doingftw'	=> 0,
		'ftwhandle'	=> undef,
		'pathto'	=> undef,
		'isdone'	=> 0
	};

	my @globtmp;
	my %nameseen;

	if ($globexp eq "" ) {
		HF::Report->report (4, 1, "FATAL globexp parameter zero-length");
	}

#	warn "globexp " . $globexp  . "\n";

	# Expand user-supplied globs
	@globtmp = glob ($globexp);
#	warn "globtmp " . join (', ', @globtmp)  . "\n";

	# Duplicate glob expansions into two entries which move together as a pair,
	# the basename to sort by and the pathname
	@globtmp = map {[ $_, $_ ]} @globtmp;
#	warn "globtmp " . join (" ", map {"(" . $_->[0] . ", " . $_->[1] . ")"} @globtmp) . "\n";

	# Strip dirnames off the basenames
	@globtmp = map {[ &HF::Misc::basepart ($_->[0]), $_->[1] ]} @globtmp;
#	warn "globtmp " . join (" ", map {"(" . $_->[0] . ", " . $_->[1] . ")"} @globtmp) . "\n";

	# Uniq the basenames.  Duplicate basename is user error, first definition of path wins
	@globtmp = grep {! $nameseen{$_->[0]}++} @globtmp;
#	warn "globtmp " . join (" ", map {"(" . $_->[0] . ", " . $_->[1] . ")"} @globtmp) . "\n";

	# Remove entries dot, dotdot, and lone slash
	@globtmp = grep {$_->[0] !~ /^(\.|\.\.|\/)$/} @globtmp;
#	warn "globtmp " . join (" ", map {"(" . $_->[0] . ", " . $_->[1] . ")"} @globtmp) . "\n";

	# Sort the basenames, don't need cmpname slash -> null treatment as paths have been removed
	@globtmp = sort {$a->[0] cmp $b->[0]} @globtmp;
#	warn "globtmp " . join (" ", map {"(" . $_->[0] . ", " . $_->[1] . ")"} @globtmp) . "\n";

	$self->{'startdir'} = &HF::Misc::dirpart ($globtmp[0]->[1]);
	$self->{'encstartdir'} = HF::Encode->encode ($self->{'startdir'});

	# Return the list of pathnames, now sorted by the pathnames' basenames
	@globtmp = map {$_->[1]} @globtmp;
#	warn "globtmp " . join (', ', @globtmp)  . "\n";

	$self->{'globresult'} = [ @globtmp ];

	bless ($self, $class);

	return ($self);
}

sub next {

	my $self = shift;

	my $nextname;
	my $res;

	if ($self->{'isdone'} == 1) {
		HF::Report->report (4, 1, "FATAL called after end of walk");
	}

	if ($self->{'firsttime'} == 1) {
		$res = { };

		$res = $self->lstat_work ($self->{'startdir'});
		$res->{'name'} = '.';
		$res->{'encname'} = '.';

		$self->{'firsttime'} = 0;
	} else {
		if ($self->{'doingftw'} == 1) {
			$res = $self->{'ftwhandle'}->next;

			if (!(defined ($res->{'inode'}))) {
				$self->{'doingftw'} = 0;
				return ($self->next);
			}

			$res->{'name'} = $self->{'pathto'} . '/' . $res->{'name'};
			$res->{'encname'} = HF::Encode->encode ($res->{'name'});

			if ((defined ($self->{'ignore'})) && ($res->{'name'} =~ /$self->{'ignore'}/)) {
				HF::Report->report (0, 7, "PROGRESS skipping '" . HF::Encode->encode ($res->{'name'}) . "'");
				if (defined ($HF::Globals::outputplbig)) {
					push (@{$HF::Globals::wasskipped}, $res->{'name'});
				}
				return ($self->next);
			}

		} else {
			$nextname = shift (@{$self->{'globresult'}});
			$res = $self->lstat_work ($nextname);

			$res->{'name'} = &HF::Misc::basepart ($res->{'name'});
			$res->{'encname'} = HF::Encode->encode ($res->{'name'});

			if ((defined ($self->{'ignore'})) && ($res->{'name'} =~ /$self->{'ignore'}/)) {
				HF::Report->report (0, 7, "PROGRESS skipping '" . HF::Encode->encode ($res->{'name'}) . "'");
				if (defined ($HF::Globals::outputplbig)) {
					push (@{$HF::Globals::wasskipped}, $res->{'name'});
				}
				return ($self->next);
			}

			if ($res->{'type'} == 0040000) {
				$self->{'pathto'} = &HF::Misc::basepart ($res->{'name'});
				$self->{'ftwhandle'} = HF::FTWstream->new ($res->{'path'}, undef, $HF::Globals::dochecksums);
				$self->{'ftwhandle'}->next;	# discard top directory
				$self->{'doingftw'} = 1;
			}
		}
	}

	if (!(defined ($res->{'inode'}))) {
		$self->{'isdone'} = 1;
	}

	return ($res);
}

sub lstat_work {

	my $self = shift;
	my $inpath = shift;

	my $res;

	$res = {
		'startdir' => $self->{'startdir'},
		'encstartdir' => HF::Encode->encode ($self->{'startdir'}),
		'path' => $inpath,
		'encpath' => HF::Encode->encode ($inpath)
	};

	($res->{'dev'}, $res->{'inode'}, $res->{'mode'},
		$res->{'links'}, $res->{'owner'}, $res->{'group'},
		$res->{'rdev'}, $res->{'size'}, $res->{'atime'},
		$res->{'mtime'}, $res->{'ctime'},
		$res->{'blksize'}, $res->{'blocks'}) = lstat ($inpath);

	if (!(defined ($res->{'inode'}))) {
		return ($res);
	} else {
		$res->{'type'} = $res->{'mode'} & 0170000;
		$res->{'perms'} = $res->{'mode'} & 07777;

		$res->{'name'} = $res->{'path'};
		if ($res->{'name'} eq '') {
			if ($res->{'type'} == 0040000) {
				$res->{'name'} = ".";
			} else {
				$res->{'name'} = $res->{'path'};
				$res->{'name'} =~ s|^.*/||;
			}
		}
		$res->{'encname'} = HF::Encode->encode ($res->{'name'});

		if ($res->{'type'} == 0120000) {
			$res->{'target'} = readlink ($inpath);
			if (!(defined ($res->{'target'}))) {
				HF::Report->report (2, 2, "FAILURE 1 can't readlink '" . $inpath . "', error '$!'", 'fail_layers');
			}
			$res->{'enctarget'} = HF::Encode->encode ($res->{'target'});

		}

		if (($res->{'type'} == 0100000) && ($self->{'dochecksums'} == 1)) {
			$res->{'checksum'} = $self->getchecksum ($inpath);
		} else {
			$res->{'checksum'} = undef;
		}

		return ($res);
	}
}

sub getroot {

	my $self = shift;

	return ($self->{'startdir'});
}

sub getchecksum {

	my $self = shift;
	my $filename = shift;

	my $MD5;

	$self->{'md5handle'}->reset;
	if (open ($MD5, "<", $filename)) {
		$self->{'md5handle'}->addfile ($MD5);
		close ($MD5) || HF::Report->report (4, 1, "FATAL close of md5 fh for filename '$filename' failed, error '$!'");
		return ($self->{'md5handle'}->hexdigest);
	} else {
		HF::Report->report (2, 2, "FAILURE 3 can't open '" . HF::Encode->encode ($filename) . "' for checksum, error '$!'", 'fail_layers');
		return (undef);
	}
}

1;
