#  Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew

package HF::STLstream;
use strict;

use Digest::MD5;

sub new {

	my $proto = shift;
	my $sl = shift;
	my $ignore = shift;
	my $dochecksums = shift;

	my $class = ref ($proto) || $proto;

	my $self = {

		'ignore'	=> $ignore,
		'dochecksums'	=> $dochecksums,

		'slfh2'	=> undef,
		'linknames'	=> { },
		'md5handle'	=> Digest::MD5->new,
		'isdone'	=> 0
	};

	my $line;
	my $seenroot;
	my @fields;
	my $theroot;
	my $slfh1;
	my $i;

	if ($sl eq "") {
		HF::Report->report (4, 1, "FATAL sl parameter zero-length");
	}

	if (!(open ($slfh1, "<", $sl))) {
		HF::Report->report (3, 1, "FATAL can't open sl file '$sl', error '$!'");
	}

	$seenroot = 0;
	$self->{'inode'} = 1000;
	while (<$slfh1>) {

		$line = $_;
		chop ($line);

		if ($line =~ /^#/) {
			next;
		}

		@fields = split (/\t/, $line);
		#  remove space padding which made numbers align in sl
		foreach $i (@fields) {
			$i =~ s/^\s+//;
		}

		if ($fields[0] eq 'root') {
			if ($seenroot == 1) {
				HF::Report->report (3, 1, "FATAL already found one root line in sl file '$sl'");
			} else {
				$self->{'startdir'} = $fields[1];
				$self->{'encstartdir'} = HF::Encode->encode ($self->{'startdir'});
				if (!(-e $self->{'startdir'})) {
					HF::Report->report (3, 1, "FATAL source walk layer $i dir '" . $self->{'encstartdir'} . "' does not exist");
				} elsif ((!(-d $self->{'startdir'})) || (-l $self->{'startdir'})) {
					HF::Report->report (3, 1, "FATAL source walk layer $i dir '" . $self->{'encstartdir'} . "' is not dir");
				}
				$seenroot = 1;
			}
		} elsif ($fields[0] eq 'hardlinked') {
			shift (@fields);
			foreach $i (@fields) {
				$self->{'linknames'}{$i} = $self->{'inode'};
			}
			$self->{'inode'}++;
		}
	}

	if (!(close ($slfh1))) {
		HF::Report->report (3, 1, "FATAL can't close sl file '$sl', error '$!'");
	}

	if ($seenroot == 0) {
		HF::Report->report (3, 1, "FATAL didn't find root line in sl file '$sl'");
	}

	if (!(open ($self->{'slfh2'}, "<", $sl))) {
		HF::Report->report (3, 1, "FATAL can't open sl file '$sl', error '$!'");
	}

	bless ($self, $class);

	return ($self);
}

sub next {

	my $self = shift;

	my $slfh2;
	my $line;
	my @fields;
	my $res;
	my $i;

	if ($self->{'isdone'} == 1) {
		HF::Report->report (4, 1, "FATAL called after end of walk");
	}

	$slfh2 = $self->{'slfh2'};
	while (<$slfh2>) {

		$line = $_;
		chop ($line);

		if ($line =~ /^#/) {
			next;
		}

		@fields = split (/\t/, $line);
		#  remove space padding which made numbers align in sl
		foreach $i (@fields) {
			$i =~ s/^\s+//;
		}

		if ($fields[0] eq 'root') {
			next;
		}
		if ($fields[0] eq 'hardlinked') {
			last;
		}
		if ((defined ($self->{'ignore'})) && ($fields[0] =~ /$self->{'ignore'}/)) {
			HF::Report->report (0, 7, "PROGRESS skipping '" . HF::Encode->encode ($fields[0]) . "'");
			if (defined ($HF::Globals::outputplbig)) {
				push (@{$HF::Globals::wasskipped}, $fields[0]);
			}
			next;
		}
		$res = {
			'startdir' => $self->{'startdir'},
			'encstartdir' => $self->{'encstartdir'},
			'encname' => $fields[9]
		};

		$res->{'name'} = HF::Encode->decode ($res->{'encname'});
		$res->{'path'} = &HF::Misc::joinpaths ($self->{'startdir'}, $res->{'name'});
		$res->{'encpath'} = HF::Encode->encode ($res->{'path'});
		$res->{'blksize'} = 1024;

		if (exists ($self->{'linknames'}{$fields[9]})) {
			$res->{'inode'} = $self->{'linknames'}{$fields[9]};
		} else {
			$res->{'inode'} = $self->{'inode'}++;
		}

		if ($fields[0] eq 'file') {
			$res->{'type'} = 0100000;
			$res->{'perms'} = oct ($fields[1]);
			$res->{'links'} = $fields[2];
			$res->{'owner'} = $fields[3];
			$res->{'group'} = $fields[4];
			$res->{'mtime'} = $fields[5];
			$res->{'size'} = $fields[6];
			$res->{'enctarget'} = '';
			$res->{'target'} = '';
			$res->{'rdev'} = 0;
			$res->{'dev'} = 999999;
			$res->{'blocks'} = $res->{'size'} / $res->{'blksize'};
			if (int ($res->{'blocks'}) < $res->{'blocks'}) {	# ceiling
				$res->{'blocks'} = int ($res->{'blocks'} + 1);
			}
			if ($self->{'dochecksums'} == 1) {
				if ($fields[10] ne '') {
					$res->{'checksum'} = $fields[10];
				} else {
					$res->{'checksum'} = $self->getchecksum ($res->{'path'});
				}
			} else {
				$res->{'checksum'} = undef;
			}
		} elsif ($fields[0] eq 'dir') {
			$res->{'type'} = 0040000;
			$res->{'perms'} = oct ($fields[1]);
			$res->{'links'} = $fields[2];
			$res->{'owner'} = $fields[3];
			$res->{'group'} = $fields[4];
			$res->{'mtime'} = $fields[5];
			$res->{'size'} = 4096;
			$res->{'enctarget'} = '';
			$res->{'target'} = '';
			$res->{'rdev'} = 0;
			$res->{'dev'} = 999999;
			$res->{'blocks'} = 4;
			$res->{'checksum'} = undef;
		} elsif ($fields[0] eq 'symlk') {
			$res->{'type'} = 0120000;
			$res->{'perms'} = 00777;
			$res->{'links'} = 1;
			$res->{'owner'} = 0;
			$res->{'group'} = 0;
			$res->{'mtime'} = 0;
			$res->{'size'} = length ($res->{'name'});
			$res->{'enctarget'} = $fields[7];
			$res->{'target'} = HF::Encode->decode ($res->{'enctarget'});
			$res->{'rdev'} = 0;
			$res->{'dev'} = 999999;
			$res->{'blocks'} = 0;
			$res->{'checksum'} = undef;
		} elsif ($fields[0] eq 'fifo') {
			$res->{'type'} = 0010000;
			$res->{'perms'} = oct ($fields[1]);
			$res->{'links'} = $fields[2];
			$res->{'owner'} = $fields[3];
			$res->{'group'} = $fields[4];
			$res->{'mtime'} = $fields[5];
			$res->{'size'} = 0;
			$res->{'enctarget'} = '';
			$res->{'target'} = '';
			$res->{'rdev'} = 0;
			$res->{'dev'} = 999999;
			$res->{'blocks'} = 0;
			$res->{'checksum'} = undef;
		} elsif ($fields[0] eq 'sock') {
			$res->{'type'} = 0140000;
			$res->{'perms'} = oct ($fields[1]);
			$res->{'links'} = $fields[2];
			$res->{'owner'} = $fields[3];
			$res->{'group'} = $fields[4];
			$res->{'mtime'} = $fields[5];
			$res->{'size'} = 0;
			$res->{'enctarget'} = '';
			$res->{'target'} = '';
			$res->{'rdev'} = 0;
			$res->{'dev'} = 999999;
			$res->{'blocks'} = 0;
			$res->{'checksum'} = undef;
		} elsif ($fields[0] eq 'block') {
			$res->{'type'} = 0060000;
			$res->{'perms'} = oct ($fields[1]);
			$res->{'links'} = $fields[2];
			$res->{'owner'} = $fields[3];
			$res->{'group'} = $fields[4];
			$res->{'mtime'} = $fields[5];
			$res->{'size'} = 0;
			$res->{'enctarget'} = '';
			$res->{'target'} = '';
			$res->{'rdev'} = $fields[8];
			$res->{'dev'} = 999999;
			$res->{'blocks'} = 0;
			$res->{'checksum'} = undef;
		} elsif ($fields[0] eq 'char') {
			$res->{'type'} = 0020000;
			$res->{'perms'} = oct ($fields[1]);
			$res->{'links'} = $fields[2];
			$res->{'owner'} = $fields[3];
			$res->{'group'} = $fields[4];
			$res->{'mtime'} = $fields[5];
			$res->{'size'} = 0;
			$res->{'enctarget'} = '';
			$res->{'target'} = '';
			$res->{'rdev'} = $fields[8];
			$res->{'dev'} = 999999;
			$res->{'blocks'} = 0;
			$res->{'checksum'} = undef;
		} else {
			HF::Report->report (3, 1, "FATAL unknown sl field type '$fields[0]'");
		}

		$res->{'mode'} = $res->{'type'} | $res->{'perms'};
		$res->{'atime'} = $res->{'mtime'};
		$res->{'ctime'} = $res->{'mtime'};

		return ($res);
	}

	if (!(close ($slfh2))) {
		HF::Report->report (3, 1, "FATAL can't close sl file, error '$!'");
	}
	$self->{'isdone'} = 1;
	$res = { };
	return ($res);
}

sub getroot {

	my $self = shift;

	return ($self->{'startdir'});
}

sub getchecksum {

	my $self = shift;
	my $filename = shift;

	my $MD5;

	$self->{'md5handle'}->reset;
	if (open ($MD5, "<", $filename)) {
		$self->{'md5handle'}->addfile ($MD5);
		close ($MD5) || HF::Report->report (4, 1, "FATAL can't close md5 fh for '$filename', error '$!'");
		return ($self->{'md5handle'}->hexdigest);
	} else {
		HF::Report->report (2, 2, "FAILURE 5 can't open '" . HF::Encode->encode ($filename) . "' for checksum, error '$!'", 'fail_layers');
		return (undef);
	}
}

1;
