#!/usr/bin/perl

#  Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew

#  oh-blocked-commit
#	oh helper to hg commit+addremove piecewise in size-limited blocks
#	blocking may prevent rename detection

use strict;

my $blocktotsize;
my $blocktotname;
my @commitexclude;
my $filecount;
my $grandtot;
my $grandtotname;
my $hgstatus;
my $line;
my $maxblock;
my $maxblockname;
my $myname;
my $omarfile;
my $partialcount;
my @sizeskippedfiles;
my @thisblock;
my $thisname;
my $thissize;

$myname = "oh-blocked-commit";

######################################################################

#  Option processing
#  No getopts, conflicts with a commit message containing a leading dash

if ($ARGV[0] eq '--version') {
	&print_version;
	exit (0);
}

if ($ARGV[0] eq '--help') {
	&print_version;
	warn ("\n");
	&print_usage;
	exit (0);
}

$maxblockname = 1_800_000;
$maxblock = 500E6;
if ($ARGV[0] eq '-b') {
	shift (@ARGV);
	$maxblock = shift (@ARGV);
}

$omarfile = shift (@ARGV);
if ($omarfile eq '') {
	&print_usage;
	exit (1);
}

######################################################################

#  main()

if (-f "$omarfile.part") {
	die "$myname: should not be file '$omarfile.part', exiting\n";
}

if (!(-f $omarfile)) {
	die "$myname: should be file '$omarfile', exiting\n";
}

if (!((-d 'oh-2.0') && (!(-d 'oh')))) {
	die "$myname: repository condition not understood, exiting\n";
}

$/ = "\0";

$filecount = 0;
$grandtot = 0;
$grandtotname = 0;
@thisblock = ( );
@sizeskippedfiles = ( );
@commitexclude = ( );
$partialcount = 0;
$blocktotsize = 0;
$blocktotname = 0;

print "STATUS doing hg commit+addremove in blocks\n";

open (HGSTATUS, 'hg status -mardu -0 oh-2.0 |') || die "$myname: hg status failed with error '$!'\n";

while (<HGSTATUS>) {

	chop;
	$line = $_;
	$hgstatus = substr ($line, 0, 1);
	$thisname = substr ($line, 2);
	$thisname =~ s|^-|./-|;

	$thissize = (lstat ($thisname))[7];
	# if file doesn't exist, then undef result is numerically zero
	if ($thissize > 750E6) {
		warn "$myname: file '$thisname' size " . &commify_number ($thissize) . " crowds 2 Gig size limit for Mercurial, skipping\n";
		push (@sizeskippedfiles, $thisname, $thissize);
		if ($hgstatus eq '?') {
			push (@commitexclude, '-X', $thisname);
		} else {
			system ('hg', 'forget', '-v', '--', $thisname);	# | cat -v
			if (($? >> 8) != 0) {
				die "$myname: hg forget failed with error '$!', quitting\n";
			}
		}
		next;
	}

	$filecount++;
	$grandtot += $thissize;
	$grandtotname += length ($thisname);

	if ((($blocktotsize + $thissize) < $maxblock) && (($blocktotname + length ($thisname)) < $maxblockname)) {
		$blocktotsize += $thissize;
		$blocktotname += length ($thisname);
		push (@thisblock, $thisname);
	} else {
		&do_commit;
		$blocktotsize = $thissize;
		$blocktotname = length ($thisname);
		@thisblock = ($thisname);
	}
}

#  Unless sizes are round numbers, there will be uncommitted files
#  left over in the partially filled block.  If omarfile pushes size
#  over, commit this partial block.  Otherwise, omit the partial
#  commit and instead let there be just a single commit.
if (($blocktotsize + (lstat ($omarfile))[7]) >= $maxblock) {
	&do_commit;
}

print "STATUS doing last hg commit+addremove\n";
if ($#ARGV > -1) {
	#  Zero-length @commitexclude gets flattened to vanish, not a null argument
	system ('hg', 'commit', '-A', '-v', @commitexclude, '-m', join (' ', @ARGV));	# | cat -v
} else {
	#  Zero-length @commitexclude gets flattened to vanish, not a null argument
	system ('hg', 'commit', '-A', '-v', @commitexclude);				# | cat -v
}
if ((($? >> 8) != 0) && (($? >> 8) != 1)) {
	die "$myname: hg commit failed with error '$!', quitting\n";
}

while ($#sizeskippedfiles > -1) {
	$thisname = shift (@sizeskippedfiles);
	$thissize = shift (@sizeskippedfiles);
	warn "$myname: file '$thisname' size " . &commify_number ($thissize) . " was too large for Mercurial, skipped it\n";
}

if ($partialcount > 0) {
	printf ("STATUS committed+addremoved %9s files in %5s partial commits plus final commit, totaling %14s bytes\n",
		&commify_number ($filecount),
		&commify_number ($partialcount),
		&commify_number ($grandtot));
} else {
	printf ("STATUS committed+addremoved %9s files, totaling %14s bytes\n",
		&commify_number ($filecount),
		&commify_number ($grandtot));
}

exit (0);

######################################################################

# Change 12345678901 into 12,345,678,901

sub do_commit {

	if ($#thisblock > -1) {
		$partialcount++;
		printf ("STATUS partial commit+addremove %5s of %9s files, size %13s of %13s, namesize %9s of %9s\n",
			&commify_number ($partialcount),
			&commify_number ($#thisblock + 1),
			&commify_number ($blocktotsize),
			&commify_number ($maxblock),
			&commify_number ($blocktotname),
			&commify_number ($maxblockname));
		system ('hg', 'commit', '-A', '-v', '-m', join (' ', @ARGV), '--', @thisblock);	# | cat -v
		if (($? >> 8) != 0) {
			die "$myname: hg commit failed with error '$!', quitting\n";
		}
	}
}

sub commify_number {

	my $number = shift;

	# In case of number like "500E6", but don't use for all
	# numbers because it will strip trailing .0
	if ($number =~ /E/) {
		$number += 0;
	}
	$number =~ s/(\d)(?=(\d{3})+(\D|$))/$1\,/g;
	return ($number);
}

sub print_version {

	warn ("omar version @@@ HFVER @@@
Copyright (C) 1994, 1995, 1996, 1997, 1998, 1999, 2002, 2003, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020 Brian Bartholomew
All rights reserved.
Written by Brian Bartholomew.
");
}

sub print_usage {

	warn ("Usage: $myname [-b maxblockbytes] omarfilename

	-b maxblockbytes
		set max change size to trigger partial commit
	--help
	--version

oh helper to hg commit+addremove piecewise in size-limited blocks
");
}
